from typing import Dict

import numpy as np
import scipy.linalg as la
import scipy.special as sp
import scipy.stats as st

from ..base import Distribution, LinkFunction, MultivariateDistributionMixin
from ..links import Identity, Log, LogShiftTwo, MatrixDiag, MatrixDiagTril
from ..types import ParameterShapes
from .mv_normal_modchol import (
    _deriv1_mu as _deriv1_mu_normal,
    _deriv2_mu as _deriv2_mu_normal,
)

# The MV gaussian distribution with modified Cholesky decomposition
# We use the notation T' D T = PRECISION
# Where T is a lower diagonal matrix and D is a diagonal matrix


class MultivariateStudentTInverseModifiedCholesky(
    MultivariateDistributionMixin, Distribution
):
    """
    Multivariate Student's t distribution with modified Cholesky decomposition.

    The PDF of the multivariate \\( t \\)-distribution with precision matrix parameterized as \\( T^T D T \\) is:
    $$
    p(y \mid \\mu, D, T, \\nu) =
    \\frac{\\Gamma\\left(\\frac{\\nu + k}{2}\\right)}
         {\\Gamma\\left(\\frac{\\nu}{2}\\right) \\left(\\pi \\nu\\right)^{k/2}}
    \\cdot \\sqrt{\\det(D)}
    \\left(1 + \\frac{1}{\\nu} (y - \\mu)^T T^T D T (y - \\mu)\\right)^{-\\frac{\\nu + k}{2}}
    $$
    where \\( k \\) is the dimensionality, \\( \\mu \\) is the location, \\( D \\) is a diagonal matrix, \\( T \\) is a lower triangular matrix, and \\( \\nu \\) is the degrees of freedom.
    """

    corresponding_gamlss: str = None
    parameter_names = {0: "mu", 1: "D", 2: "T", 3: "dof"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (-np.inf, np.inf),
        3: (1, np.inf),
    }
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.multivariate_normal
    scipy_names = {"mu": "loc", "sigma": "shape", "nu": "df"}
    parameter_shape = {
        0: ParameterShapes.VECTOR,
        1: ParameterShapes.DIAGONAL_MATRIX,
        2: ParameterShapes.LOWER_TRIANGULAR_MATRIX,
        3: ParameterShapes.SCALAR,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link_1: LinkFunction = MatrixDiag(Log()),
        scale_link_2: LinkFunction = MatrixDiagTril(Identity(), Identity()),
        tail_link: LinkFunction = LogShiftTwo(),
        dof_guesstimate: float = 10,
        use_gaussian_for_location: bool = False,
    ):
        """
        Initializes the distribution with specified link functions and parameters.
        Args:
            loc_link (LinkFunction, optional): Link function for the location parameter. Defaults to Identity().
            scale_link_1 (LinkFunction, optional): Link function for the first scale parameter (diagonal). Defaults to MatrixDiag(Log()).
            scale_link_2 (LinkFunction, optional): Link function for the second scale parameter (lower-triangular). Defaults to MatrixDiagTril(Identity(), Identity()).
            tail_link (LinkFunction, optional): Link function for the tail (degrees of freedom) parameter. Defaults to LogShiftTwo().
            dof_guesstimate (float, optional): Initial guess for the degrees of freedom. Defaults to 10.
            use_gaussian_for_location (bool, optional): Whether to use a Gaussian approximation for the location parameter. Defaults to False.
        Attributes:
            dof_guesstimate (float): Stores the initial guess for the degrees of freedom.
            dof_independence (float): Large value to represent independence in degrees of freedom.
            use_gaussian_for_location (bool): Indicates if a Gaussian is used for the location parameter.
            _regularization_allowed (dict): Specifies which parameters allow regularization.
        """
        super().__init__(
            links={
                0: loc_link,
                1: scale_link_1,
                2: scale_link_2,
                3: tail_link,
            }
        )
        self.dof_guesstimate = dof_guesstimate
        self.dof_independence = 1e6
        self.use_gaussian_for_location = use_gaussian_for_location
        self._regularization_allowed = {0: False, 1: False, 2: True, 3: False}

    def get_adr_regularization_distance(self, dim: int, param: int) -> float:
        if param in (0, 1, 3):
            return None
        if param == 2:
            i, j = np.tril_indices(dim, k=-1)
            return np.abs(i - j)

    def get_regularization_size(self, dim: int) -> float:
        return dim - 1

    @staticmethod
    def fitted_elements(dim: int):
        return {0: dim, 1: dim, 2: int(dim * (dim - 1) // 2), 3: 1}

    @staticmethod
    def index_flat_to_cube(k: int, d: int, param: int):
        if param in (0, 3):
            return k
        if param == 1:
            return k, k
        if param == 2:
            # tril_indicies is row-wise
            # "inverted" triu_indicies is column-wise
            i, j = np.tril_indices(d, k=-1)
            return i[k], j[k]

    def set_theta_element(
        self, theta: Dict, value: np.ndarray, param: int, k: int
    ) -> Dict:
        """Sets an element of theta for parameter param and place k.

        !!! Note
            This will mutate `theta`!

        Args:
            theta (Dict): Current fitted $\theta$
            value (np.ndarray): Value to set
            param (int): Distribution parameter
            k (int): Flat element index $k$

        Returns:
            Dict: Theta where element (param, k) is set to value.
        """
        if param in (0, 3):
            theta[param][:, k] = value
        if param == 1:
            theta[param][:, k, k] = value
        if param == 2:
            d = theta[0].shape[1]
            i, j = self.index_flat_to_cube(k, d, param)
            theta[param][:, i, j] = value
        return theta

    def theta_to_params(self, theta):
        mu = theta[0]
        d_mat = theta[1]
        t_mat = theta[2]
        dof = theta[3]
        return mu, d_mat, t_mat, dof

    def set_initial_guess(
        self,
        y: np.ndarray,
        theta: Dict[int, np.ndarray],
        param: int,
    ) -> Dict[int, np.ndarray]:
        if param in (0, 2):
            return theta
        if param == 1:
            M = y.shape[0]
            residual = y - theta[0]
            var = np.var(residual, axis=0)
            shape = np.diag(var * (self.dof_guesstimate - 2) / self.dof_guesstimate)
            _, d_inv, _ = la.ldl(shape, lower=True)
            d_mat = np.linalg.inv(d_inv)
            theta[1] = np.tile(d_mat, (M, 1, 1))
            return theta
        if param == 3:
            theta[3] = np.full_like(theta[3], self.dof_guesstimate)
            return theta

    def dl1_dp1(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dp2(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dpp(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented.")

    def element_score(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl1_dp1(y=y, theta=theta, param=param, k=k)

    def element_hessian(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl2_dp2(y=y, theta=theta, param=param, k=k)

    def element_dl1_dp1(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip=False
    ):
        mu, d_mat, t_mat, dof = self.theta_to_params(theta)
        if (param == 0) and not self.use_gaussian_for_location:
            deriv = _deriv1_mu(y, mu, d_mat, t_mat, dof, i=k)
        elif (param == 0) and self.use_gaussian_for_location:
            deriv = _deriv1_mu_normal(y, mu, d_mat, t_mat, i=k)
        if param == 1:
            deriv = _deriv1_dmat(y, mu, d_mat, t_mat, dof, i=k)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            deriv = _deriv1_tmat(y, mu, d_mat, t_mat, dof, i=i, j=j)
        if param == 3:
            deriv = _deriv1_dof(y, mu, d_mat, t_mat, dof)
        return deriv

    def element_dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip=False
    ):
        mu, d_mat, t_mat, dof = self.theta_to_params(theta)
        if (param == 0) and not self.use_gaussian_for_location:
            deriv = _deriv2_mu(y, mu, d_mat, t_mat, dof, i=k)
        if (param == 0) and self.use_gaussian_for_location:
            deriv = _deriv2_mu_normal(y, mu, d_mat, t_mat, i=k)
        if param == 1:
            deriv = _deriv2_dmat(y, mu, d_mat, t_mat, dof, i=k)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            deriv = _deriv2_tmat(y, mu, d_mat, t_mat, dof, i=i, j=j)
        if param == 3:
            deriv = _deriv2_dof(y, mu, d_mat, t_mat, dof)
        return deriv

    def element_link_function(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param in (0, 3):
            return self.links[param].link(y)
        if param in (1, 2):
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_link(y, i=i, j=j)

    def element_link_function_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param in (0, 3):
            return self.links[param].link_derivative(y)
        if param in (1, 2):
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_derivative(y, i=i, j=j)

    def element_link_function_second_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param in (0, 3):
            return self.links[param].link_second_derivative(y)
        if param in (1, 2):
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_link_second_derivative(y, i=i, j=j)

    def element_link_inverse(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param in (0, 3):
            return self.links[param].inverse(y)
        if param in (1, 2):
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_inverse(y, i=i, j=j)

    def element_link_inverse_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param in (0, 3):
            return self.links[param].inverse_derivative(y)
        if param in (1, 2):
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_inverse_derivative(y, i=i, j=j)

    def initial_values(self, y, param=0):
        M = y.shape[0]
        mu = np.mean(y, axis=0)
        var = np.var(y, axis=0)
        shape = np.diag(var * (self.dof_guesstimate - 2) / self.dof_guesstimate)
        t_inv, d_inv, _ = la.ldl(shape, lower=True)
        t_mat = np.linalg.inv(t_inv)
        d_mat = np.linalg.inv(d_inv)

        if param == 0:
            return np.tile(mu, (M, 1))
        if param == 1:
            return np.tile(d_mat, (M, 1, 1))
        if param == 2:
            return np.tile(t_mat, (M, 1, 1))
        if param == 3:
            return np.full((M, 1), self.dof_independence)

    def cube_to_flat(self, x: np.ndarray, param: int):
        if param in (0, 3):
            return x
        if param == 1:
            k = x.shape[1]
            return x[:, range(k), range(k)]
        if param == 2:
            k = x.shape[1]
            # column wise indices
            i, j = np.tril_indices(k, k=-1)
            return x[:, i, j]

    def flat_to_cube(self, x: np.ndarray, param: int):
        if param in (0, 3):
            return x
        if param == 1:
            k = x.shape[1]
            out = np.zeros((x.shape[0], k, k))
            out[:, range(k), range(k)] = x
            return out

        if param == 2:
            n, k = x.shape
            # The following conversion holds for upper diagonal matrices
            # We INCLUDE the diagonal!!
            # (D + 1) * D // 2 = k
            # (D + 1) * D = 2 * k
            # D^2 + D = 2 * k
            # ... Wolfram Alpha ...
            # D = 0.5 * (sqrt(8k + 1) - 1)
            d = int(1 / 2 * (np.sqrt(8 * k + 1) - 1))
            i, j = np.tril_indices(d + 1, k=-1)
            out = np.zeros((n, d + 1, d + 1))
            out[:, range(d + 1), range(d + 1)] = 1
            out[:, i, j] = x
            return out

    def param_conditional_likelihood(
        self, y: np.ndarray, theta: Dict, eta: np.ndarray, param: int
    ) -> np.ndarray:
        """Calulate the log-likelihood for (flat) eta for parameter (param)
        and theta for all other parameters.

        Args:
            y (np.ndarray): True values
            theta (Dict): Fitted theta.
            eta (np.ndarray): Fitted eta.
            param (int): Param for which we take eta.

        Returns:
            np.ndarray: Log-likelihood.
        """
        fitted = self.flat_to_cube(eta, param=param)
        fitted = self.link_inverse(fitted, param=param)
        # fitted_theta = {**theta, param: fitted_eta}
        return self.logpdf(y, theta={**theta, param: fitted})

    def theta_to_scipy(self, theta: Dict[int, np.ndarray]):
        mu, d_mat, t_mat, dof = self.theta_to_params(theta)
        cov = np.linalg.inv(t_mat.swapaxes(-1, -2) @ d_mat @ t_mat)
        out = {
            "loc": mu,
            "shape": cov,
            "df": dof,
        }
        return out

    def cdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def ppf(self, q, theta):
        raise NotImplementedError("Not implemented")

    def rvs(self, size, theta):
        raise NotImplementedError("Not implemented")

    def logcdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def logpdf(self, y, theta):
        return _loglikelihood_t_modchol(y, theta[0], theta[1], theta[2], theta[3])

    def logpmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def calculate_conditional_initial_values(
        self, y: np.ndarray, theta: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        raise NotImplementedError("Not implemented")


def _loglikelihood_t_modchol(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
) -> Dict[str, float]:
    k = y.shape[1]
    y_centered = y - mu
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    A = np.squeeze(sp.gammaln((dof + k) / 2))
    B = np.squeeze(sp.gammaln(dof / 2))
    C = 1 / 2 * k * np.squeeze(np.log(np.pi * dof))
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)
    part1 = A - B - C
    # part2 = 1 / 2 * np.log(np.linalg.det(precision))
    part2 = 1 / 2 * np.log(np.linalg.det(D_inv))
    part3 = np.squeeze((dof + k) / 2) * np.log((1 + np.squeeze((1 / dof)) * Z))
    return part1 + part2 - part3


def _deriv1_mu(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)
    part1 = (k + dof.squeeze()) / (2 * (Z + dof.squeeze()))
    part2 = 2 * np.sum(precision[:, i, :] * y_centered, axis=1)
    return part1 * part2


def _deriv2_mu(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)
    deriv_1 = 2 * np.sum(precision[:, i, :] * y_centered, axis=1)
    deriv_2 = 2 * precision[:, i, i]
    part1 = k + dof.squeeze()
    part2 = (Z + dof.squeeze()) * deriv_2 - deriv_1**2
    part3 = 2 * (Z + dof.squeeze()) ** 2
    return -(part1 * part2) / part3


def _deriv1_dmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)

    deriv_logdet = 1 / 2 * 1 / D_inv[:, i, i]

    part2 = np.squeeze(k + dof) / (2 * (Z + dof.squeeze()))
    part3 = Ty[:, i] ** 2
    deriv_Z = part2 * part3
    return deriv_logdet - deriv_Z


def _deriv2_dmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)

    # Derivative of logdet part
    part1 = -1 / 2 * 1 / D_inv[:, i, i] ** 2

    # Derivatives of Z
    deriv1 = Ty[:, i] ** 2
    deriv2 = 0

    part2 = k + dof.squeeze()
    part3 = (Z + dof.squeeze()) * deriv2 - deriv1**2
    part4 = 2 * (Z + dof.squeeze()) ** 2
    return part1 - (part2 * part3) / part4


def _deriv1_tmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
    i: int = 0,
    j: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)

    part1 = 0  # logdet part
    part2 = np.squeeze(k + dof) / (2 * (Z + dof.squeeze()))
    part3 = 2 * D_inv[:, i, i] * Ty[:, i] * y_centered[:, j]

    return part1 - part2 * part3


def _deriv2_tmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
    i: int = 0,
    j: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)

    part1 = 0  # Derivative of logdet part
    # first and second
    # deriv_Z = (-1 / 2) * 2 * D_inv[:, i, i] * Ty[:, i] * y_centered[:, j]
    # deriv_Z = (-1 / 2) * 2 * D_inv[:, i, i] * y_centered[:, j] ** 2
    deriv1 = 2 * D_inv[:, i, i] * Ty[:, i] * y_centered[:, j]
    deriv2 = 2 * D_inv[:, i, i] * y_centered[:, j] ** 2

    part2 = k + dof.squeeze()
    part3 = (Z + dof.squeeze()) * deriv2 - deriv1**2
    part4 = 2 * (Z + dof.squeeze()) ** 2
    return part1 - (part2 * part3) / part4


def _deriv1_dof(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
):
    k = y.shape[1]
    y_centered = y - mu
    dof = np.squeeze(dof)
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)

    part1 = -(-dof * sp.digamma((k + dof) / 2) + k + dof * sp.digamma(dof / 2)) / (
        2 * dof
    )
    part2 = (Z * (k + dof)) / (2 * dof * (dof + Z)) - 1 / 2 * np.log((dof + Z) / dof)
    return part1 + part2


def _deriv2_dof(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    dof: np.ndarray,
):
    k = y.shape[1]
    y_centered = y - mu
    dof = np.squeeze(dof)
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)
    part1 = (
        1
        / 4
        * (
            (2 * k) / (dof**2)
            + sp.polygamma(1, (dof + k) / 2)
            - sp.polygamma(1, dof / 2)
        )
    )
    part2 = (Z * (dof * Z - k * (2 * dof + Z))) / (2 * dof**2 * (dof + Z) ** 2)
    return part1 + part2



from typing import Tuple

import numpy as np
from scipy.special import digamma, gamma, gammaln, polygamma
from scipy.stats import gamma as gamma_dist

from ..base import Distribution, LinkFunction
from ..links import Identity, Log
from ..types import ParameterShapes


class PowerExponential(Distribution):
    """
    Power Exponential distribution (GAMLSS: PE).

    The PDF is defined as:

    \\(f(y \\mid \\mu, \\sigma, \\nu) = \\left(\\frac{\\nu}{2 \\sigma c(\\nu)}\\right) \\exp\\left(-0.5 \\left|\\frac{y - \\mu}{\\sigma c(\\nu)}\\right|^{\\nu}\\right)\\)

    where \\(c(\\nu) = \\sqrt{2} \\left(\\frac{\\Gamma(1/\\nu)}{\\Gamma(3/\\nu)}\\right)^{1/2}\\) and \\(\\Gamma(\\cdot)\\) is the gamma function.

    The parameters are: $\\mu$ (location), $\\sigma$ (scale), and $\\nu$ (shape).


    """

    corresponding_gamlss = "PE"
    parameter_names = {0: "mu", 1: "sigma", 2: "nu"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (-np.inf, np.inf)
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
        2: ParameterShapes.SCALAR,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
        shape_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link, 2: shape_link})

    def _log_c(self, nu):
        return 0.5 * (-(2 / nu) * np.log(2) + gammaln(1 / nu) - gammaln(3 / nu))

    def _c(self, nu):
        return np.exp(self._log_c(nu))

    def _z(self, y, mu, sigma):
        return (y - mu) / sigma

    def logpdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        c = self._c(nu)
        z = self._z(y, mu, sigma)
        return (
            np.log(nu)
            - np.log(sigma)
            - self._log_c(nu)
            - 0.5 * np.abs(z / c) ** nu
            - (1 + 1 / nu) * np.log(2)
            - gammaln(1 / nu)
        )

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def cdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        c = self._c(nu)
        z = self._z(y, mu, sigma)
        s = 0.5 * (np.abs(z / c) ** nu)
        sign_z = np.sign(z)
        return 0.5 * (1 + gamma_dist.cdf(s, a=1 / nu, scale=1) * sign_z)

    def logcdf(self, y, theta):
        return np.log(self.cdf(y, theta))

    def ppf(self, q, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        c = self._c(nu)

        q = np.asarray(q)

        # Output container
        result = np.full_like(q, fill_value=np.nan, dtype=np.float64)

        # Handling boundary conditions explicitly
        result[q == 0] = -np.inf
        result[q == 1] = np.inf

        # Validating quantiles between 0 and 1 (excluding boundaries)
        valid = (q > 0) & (q < 1)
        p = q[valid]
        sign = np.sign(p - 0.5)
        gamma_q = gamma_dist.ppf(np.abs(2 * p - 1), a=1 / nu, scale=1)
        z = sign * (2 * gamma_q) ** (1 / nu) * c
        result[valid] = mu + sigma * z

        return result

    def calculate_conditional_initial_values(self, y, theta, param):
        if param == 0:
            # Initial guess for mu
            return np.full_like(y, np.mean(y))
        elif param == 1:
            # Initial guess for sigma
            return np.full_like(y, np.std(y))
        elif param == 2:
            # Initial guess for nu
            return np.full_like(y, 2.0)
        else:
            raise ValueError("Invalid parameter index.")

    def rvs(self, theta, n=1, random_state=None):
        rng = np.random.default_rng(random_state)
        p = rng.uniform(size=n)
        return self.ppf(p, theta)

    def dl1_dp1(self, y, theta, param):
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)
        z = self._z(y, mu, sigma)
        c = self._c(nu)

        if param == 0:
            return (np.sign(z) * nu / (2 * sigma * np.abs(z))) * (np.abs(z / c) ** nu)
        elif param == 1:
            return ((nu / 2) * (np.abs(z / c) ** nu) - 1) / sigma
        elif param == 2:
            logc_deriv = (1 / (2 * nu**2)) * (
                2 * np.log(2) - digamma(1 / nu) + 3 * digamma(3 / nu)
            )
            s = np.abs(z / c) ** nu
            return (
                (1 / nu)
                - 0.5 * np.log(np.abs(z / c)) * s
                + np.log(2) / (nu**2)
                + digamma(1 / nu) / (nu**2)
                + (-1 + (nu / 2) * s) * logc_deriv
            )

    def d2ldv2(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)

        dlogc_dv = (1 / (2 * nu**2)) * (
            2 * np.log(2) - digamma(1 / nu) + 3 * digamma(3 / nu)
        )

        p = (1 + nu) / nu
        part1 = p * polygamma(1, p) + 2 * digamma(p) ** 2
        part2 = digamma(p) * (np.log(2) + 3 - 3 * digamma(3 / nu) - nu)
        part3 = -3 * digamma(3 / nu) * (1 + np.log(2))
        part4 = -(nu + np.log(2)) * np.log(2)
        part5 = -nu + (nu**4) * (dlogc_dv**2)

        d2ldv2 = -(part1 + part2 + part3 + part4 + part5) / (nu**3)
        d2ldv2 = np.where(d2ldv2 < -1e-15, d2ldv2, -1e-15)
        return d2ldv2

    def dl2_dp2(self, y, theta, param):
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)

        if param == 0:
            z = self._z(y, mu, sigma)
            c = self._c(nu)
            dldm = (np.sign(z) * nu / (2 * sigma * np.abs(z))) * (np.abs(z / c) ** nu)
            closed_form = -(nu**2 * gamma(2 - 1 / nu) * gamma(3 / nu)) / (
                (sigma * gamma(1 / nu)) ** 2
            )
            d2ldm2 = np.where(nu < 1.05, -(dldm**2), closed_form)
            return d2ldm2

        if param == 1:
            return -nu / (sigma**2)

        if param == 2:
            return self.d2ldv2(y, theta)

    def dl2_dpp(self, y, theta, params):
        self._validate_dl2_dpp_inputs(y, theta, params)
        if (set(params) == {0, 1}) or (set(params) == {0, 2}):
            return np.zeros_like(y)

        _, sigma, nu = self.theta_to_params(theta)
        return (1 / (2 * sigma)) * (
            (3 / nu) * (digamma(1 / nu) - digamma(3 / nu)) + 2 + (2 / nu)
        )

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        mu_init = np.full_like(y, np.mean(y))
        sigma_init = np.full_like(y, np.std(y, ddof=1))
        nu_init = np.full_like(y, np.log(2))
        return np.stack((mu_init, sigma_init, nu_init), axis=1)

    def theta_to_params(self, theta: np.ndarray) -> Tuple[np.ndarray, ...]:
        theta = np.atleast_2d(theta)  # This ensures it's always 2D
        return tuple(theta[:, i] for i in range(self.n_params))

    def pmf(self, *args, **kwargs):
        raise NotImplementedError("PMF is not defined for continuous distributions")

    def logpmf(self, *args, **kwargs):
        raise NotImplementedError("logPMF is not defined for continuous distributions")



from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Logit
from ..types import ParameterShapes


class Beta(ScipyMixin, Distribution):
    """The Beta Distribution for GAMLSS.

    The distribution function is defined as in GAMLSS as:
    $$
    f(y|\\mu,\\sigma)=\\frac{\\Gamma(\\frac{1 - \sigma^2}{\sigma^2})}
        {
        \\Gamma(\\frac{\\mu (1 - \\sigma^2)}{\\sigma^2})
        \Gamma(\\frac{(1 - \\mu) (1 - \sigma^2)}{\\sigma^2})}
        y^{\\frac{\mu (1 - \\sigma^2)}{\\sigma^2} - 1}
        (1-y)^{\\frac{(1 - \\mu) (1 - \\sigma^2)}{\\sigma^2} - 1}
    $$



    with the location and shape parameters $\mu, \sigma > 0$.

    !!! Note
        The function is parameterized as GAMLSS' BE() distribution.

        This parameterization is different to the `scipy.stats.beta(alpha, beta, loc, scale)` parameterization.

        We can use `Beta().gamlss_to_scipy(mu, sigma)` to map the distribution parameters to scipy.

    The `scipy.stats.beta()` distribution is defined as:
    $$
    f(x, \\alpha, \\beta) = \\frac{\\Gamma(\\alpha + \\beta) x^{\\alpha - 1} {(1 - x)}^{\\beta - 1}}{\Gamma(\\alpha) \\Gamma(\\beta)}
    $$

    with the paramters $\\alpha, \\beta >0$. The parameters can be mapped as follows:
    $$
    \\alpha = \\mu (1 - \\sigma^2) / \\sigma^2 \\Leftrightarrow \\mu = \\alpha / (\\alpha + \\beta)
    $$
    and
    $$
    \\beta = (1 - \\mu) (1 - \\sigma^2)/ \\sigma^2 \\Leftrightarrow \\sigma = \\sqrt{((\\alpha + \\beta + 1) )}
    $$


    Args:
        loc_link (LinkFunction, optional): The link function for $\mu$. Defaults to  LOGIT
        scale_link (LinkFunction, optional): The link function for $\sigma$. Defaults to LOGIT
    """

    corresponding_gamlss: str = "BE"

    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.nextafter(1, 0)),
        1: (np.nextafter(0, 1), np.nextafter(1, 0)),
    }
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
    }
    distribution_support = (np.nextafter(0, 1), np.nextafter(1, 0))
    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.beta
    # Theta columns do not map 1:1 to scipy parameters for beta
    # So we have to overload theta_to_scipy_params
    scipy_names = {}

    def __init__(
        self,
        loc_link: LinkFunction = Logit(),
        scale_link: LinkFunction = Logit(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link})

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        """Map GAMLSS Parameters to scipy parameters.

        Args:
            theta (np.ndarray): parameters

        Returns:
            dict: Dict of (a, b, loc, scale) for scipy.stats.beta(a, b, loc, scale)
        """
        mu = theta[:, 0]
        sigma = theta[:, 1]
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2
        params = {"a": alpha, "b": beta, "loc": 0, "scale": 1}
        return params

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return ((1 - sigma**2) / sigma**2) * (
                -spc.digamma(alpha) + spc.digamma(beta) + np.log(y) - np.log(1 - y)
            )

        if param == 1:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return -(2 / sigma**3) * (
                mu * (-spc.digamma(alpha) + spc.digamma(alpha + beta) + np.log(y))
                + (1 - mu)
                * (-spc.digamma(beta) + spc.digamma(alpha + beta) + np.log(1 - y))
            )

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            # MU
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return -(((1 - sigma**2) ** 2) / sigma**4) * (
                spc.polygamma(1, alpha) + spc.polygamma(1, beta)
            )

        if param == 1:
            # SIGMA
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return -(4 / (sigma**6)) * (
                (mu**2) * spc.polygamma(1, alpha)
                + ((1 - mu) ** 2) * spc.polygamma(1, beta)
                - spc.polygamma(1, alpha + beta)
            )

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma = self.theta_to_params(theta)

        if sorted(params) == [0, 1]:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return (2 * (1 - sigma**2) / sigma**5) * (
                mu * spc.polygamma(1, alpha) - (1 - mu) * spc.polygamma(1, beta)
            )

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y, axis=0), 0.5]
        return np.tile(initial_params, (y.shape[0], 1))



import sys

import numpy as np


def calculate_asymptotic_training_length(forget: float):
    if forget == 0:
        # Return the highest possible value that is no infinity
        return sys.maxsize
    else:
        return 1 / forget


def calculate_effective_training_length(forget: float, n_obs: int):
    if forget == 0:
        return n_obs
    else:
        return (1 - (1 - forget) ** n_obs) / forget


def _online_mean_update(avg: float, value: float, forget: float, n_seen: int):
    n_asymmptotic = calculate_asymptotic_training_length(forget)
    n_eff = calculate_effective_training_length(forget, n_seen)
    forget_scaled = forget * np.maximum(n_asymmptotic / n_eff, 1.0)
    diff = value - avg
    incr = forget_scaled * diff
    if forget_scaled > 0:
        new_avg = avg + incr
    else:
        new_avg = avg + diff / n_seen
    return new_avg


def online_mean_update(
    avg: float,
    value: float | np.ndarray,
    forget: float,
    n_seen: int,
) -> float:
    """Update the average with a new value or an array of values using an online mean update.

    Args:
        avg (float): The current average.
        value (float | np.ndarray): The new value or array of values to update the average with.
            If `value` is an array, the average is updated iteratively for each element.
        forget (float): The forgetting factor.
        n_seen (int): The number of observations seen so far.

    Returns:
        float: The updated average.
    """
    if isinstance(value, np.ndarray):
        _avg = float(avg)
        for i, v in zip(range(-value.shape[0] + 1, 1), value, strict=True):
            # Substract since we expect n_seen to be number of observations INCLUDING
            # the ones in values
            _avg = _online_mean_update(_avg, v, forget, n_seen + i)
        return _avg
    else:
        return _online_mean_update(avg, value, forget, n_seen)



from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction
from ..links import Log, Logit
from ..types import ParameterShapes


class BetaInflated(Distribution):
    """The Beta Inflated Distribution for GAMLSS.
    
    The distribution function is defined as in GAMLSS as:
    $$
    f_Y(y \\mid \\mu, \\sigma, \\nu, \\tau) = 
    \\begin{cases}
    p_0 & \\text{if } y = 0 \\
    (1 - p_0 - p_1) \\dfrac{1}{B(\\alpha, \\beta)} y^{\\alpha - 1}(1 - y)^{\\beta - 1} & \\text{if } 0 < y < 1 \\
    p_1 & \\text{if } y = 1
    \\end{cases}
    $$
       
    where $\\alpha = \\mu (1 - \\sigma^2) / \\sigma^2$, \\beta = (1 - \\mu) (1 - \\sigma^2)/ \\sigma^2; 
    p_0 = \\nu (1 + \\nu + \\tau)^{-1} and p_1 =  \\tau (1 + \\nu + \\tau)^{-1}$, 

    and $\\mu, \\sigma \\in (0,1)$ and $\\nu, \\tau > 0 $

    The parameter tuple $\\theta$ in Python is defined as:

    $\\theta = (\\theta_0, \\theta_1, \\theta_2, \\theta_3) = (\\mu, \\sigma, \\nu, \\tau)$ 
    where $\\mu = \\theta_0$ is the location parameter, $\\sigma = \\theta_1$ is the scale parameter 
    and $\\nu, \\tau = \\theta_2, \\theta_3$ are shape parameters which together define the inflation at 0 and 1

    This distribution corresponds to the BEINF() distribution in GAMLSS.
    """

    corresponding_gamlss: str = "BEINF"

    parameter_names = {0: "mu", 1: "sigma", 2: "nu", 3: "tau"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.nextafter(1, 0)),
        1: (np.nextafter(0, 1), np.nextafter(1, 0)),
        2: (np.nextafter(0, 1), np.inf),  ##
        3: (np.nextafter(0, 1), np.inf),  ##
    }
    distribution_support = (0, 1)
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
        2: ParameterShapes.SCALAR,
        3: ParameterShapes.SCALAR,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Logit(),
        scale_link: LinkFunction = Logit(),
        skew_link: LinkFunction = Log(),  ##
        tail_link: LinkFunction = Log(),  ##
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link, 2: skew_link, 3: tail_link})

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu, tau = self.theta_to_params(theta)

        if param == 0:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = ((1 - sigma**2) / sigma**2) * (
                -spc.digamma(alpha) + spc.digamma(beta) + np.log(y) - np.log(1 - y)
            )

            return np.where((y == 0) | (y == 1), 0, result)

        if param == 1:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = -(2 / sigma**3) * (
                mu * (-spc.digamma(alpha) + spc.digamma(alpha + beta) + np.log(y))
                + (1 - mu)
                * (-spc.digamma(beta) + spc.digamma(alpha + beta) + np.log(1 - y))
            )

            return np.where((y == 0) | (y == 1), 0, result)

        if param == 2:
            return np.where(
                y == 0, (1 / nu) - (1 / (1 + nu + tau)), 0 - (1 / (1 + nu + tau))
            )

        if param == 3:
            return np.where(
                y == 1, (1 / tau) - (1 / (1 + nu + tau)), 0 - (1 / (1 + nu + tau))
            )

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu, tau = self.theta_to_params(theta)
        if param == 0:
            # MU
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = -(((1 - sigma**2) ** 2) / sigma**4) * (
                spc.polygamma(1, alpha) + spc.polygamma(1, beta)
            )

            return np.where((y == 0) | (y == 1), 0, result)

        if param == 1:
            # SIGMA
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = -(4 / (sigma**6)) * (
                (mu**2) * spc.polygamma(1, alpha)
                + ((1 - mu) ** 2) * spc.polygamma(1, beta)
                - spc.polygamma(1, alpha + beta)
            )

            return np.where((y == 0) | (y == 1), 0, result)

        if param == 2:
            return -(1 + tau) / (nu * ((1 + nu + tau) ** 2))

        if param == 3:
            return -(1 + nu) / (tau * ((1 + nu + tau) ** 2))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma, nu, tau = self.theta_to_params(theta)

        if sorted(params) == [0, 1]:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = (2 * (1 - sigma**2) / sigma**5) * (
                mu * spc.polygamma(1, alpha) - (1 - mu) * spc.polygamma(1, beta)
            )

            return np.where((y == 0) | (y == 1), 0, result)

        if sorted(params) == [0, 2]:
            return np.zeros_like(y)  ###

        if sorted(params) == [0, 3]:
            return np.zeros_like(y)  ###

        if sorted(params) == [1, 2]:
            return np.zeros_like(y)  ###

        if sorted(params) == [1, 3]:
            return np.zeros_like(y)  ###

        if sorted(params) == [2, 3]:
            return 1 / (1 + nu + tau) ** 2  ###

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        return np.tile([np.mean(y), 0.5, 5, 5], (y.shape[0], 1))

    def cdf(self, y, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = alpha * (1 - mu) / mu

        cdf_beta = st.beta(alpha, beta).cdf(y)

        raw_cdf = np.where(
            (y > 0) & (y < 1),
            nu + cdf_beta,
            np.where(y == 0, nu, np.where(y == 1, 1 + nu + tau, 0)),
        )

        result = raw_cdf / (1 + nu + tau)

        result = np.where(y < 0, 0, result)
        result = np.where(y > 1, 1, result)

        return result

    def pdf(self, y, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        pdf_beta = st.beta(alpha, beta, loc=0, scale=1).pdf(y)

        result = np.where(
            (y < 0) | (y > 1),
            0,
            np.where(
                y == 0,
                nu / (1 + nu + tau),
                np.where(y == 1, tau / (1 + nu + tau), pdf_beta / (1 + nu + tau)),
            ),
        )
        return result

    def ppf(self, q, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)

        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        denom = 1 + nu + tau
        lower = nu / denom
        upper = (1 + nu) / denom

        q = np.asarray(q, dtype=np.float64)
        adjusted_q = (q - lower) * denom  # equivalent to (q - lower) / (1/denom)

        finite_result = st.beta.ppf(adjusted_q, a=alpha, b=beta)

        one_result = np.where(np.logical_or(q == 1, q >= upper), 1.0, finite_result)
        zero_result = np.where(np.logical_or(q == 0, q <= lower), 0.0, one_result)
        result = np.where(np.logical_or(q < 0, q > 1), np.nan, zero_result)

        return result

    def pmf(self, y, theta):
        raise NotImplementedError("PMF is not implemented for mixed distributions.")

    def rvs(self, size, theta):
        sim_unif = st.uniform.rvs(size=size)
        return self.ppf(q=sim_unif, theta=theta)

    def logpdf(self, y, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        logpdf_beta = st.beta(alpha, beta, loc=0, scale=1).logpdf(y)

        result = np.where(
            (y < 0) | (y > 1),
            0,
            np.where(
                y == 0,
                np.log(nu) - np.log(1 + nu + tau),
                np.where(
                    y == 1,
                    np.log(tau) - np.log(1 + nu + tau),
                    logpdf_beta - np.log(1 + nu + tau),
                ),
            ),
        )
        return result

    def logcdf(self, y, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        y = np.asarray(y, dtype=np.float64)
        denom = 1 + nu + tau

        cdf_beta = st.beta(alpha, beta).cdf(y)

        raw_cdf = np.where(
            (y > 0) & (y < 1),
            nu + cdf_beta,
            np.where(y == 0, nu, np.where(y == 1, denom, 0)),
        )

        with np.errstate(divide="ignore"):
            log_result = np.log(raw_cdf / denom)

        log_result = np.where(y < 0, -np.inf, log_result)
        log_result = np.where(y > 1, 0.0, log_result)

        return log_result

    def logpmf(self, y, theta):
        return super().logpmf(y, theta)

    def calculate_conditional_initial_values(self, y, theta, param):
        return super().calculate_conditional_initial_values(y, theta, param)



from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log, LogShiftTwo
from ..types import ParameterShapes


class SkewT(Distribution):
    corresponding_gamlss: str = "ST3"
    parameter_names = {0: "mu", 1: "sigma", 2: "nu", 3: "tau"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (np.nextafter(0, 1), np.inf),
        3: (np.nextafter(0, 1), np.inf),
    }
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
        2: ParameterShapes.SCALAR,
        3: ParameterShapes.SCALAR,
    }
    distribution_support = (-np.inf, np.inf)
    scipy_dist = None
    scipy_names = None

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
        skew_link: LinkFunction = Log(),
        shape_link: LinkFunction = LogShiftTwo(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
                2: skew_link,
                3: shape_link,
            }
        )

    def cdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        mu, sigma, nu, tau = self.theta_to_params(theta)
        cdf1 = 2 * st.t(df=tau, loc=0, scale=1).cdf(nu * (y - mu) / sigma)
        cdf2 = 1 + 2 * nu * nu * (
            st.t(df=tau, loc=0, scale=1).cdf((y - mu) / (sigma * nu)) - 0.5
        )
        cdf = np.where(y < mu, cdf1, cdf2) / (1 + nu**2)
        return cdf

    def ppf(self, q: np.ndarray, theta: np.ndarray) -> np.ndarray:
        mu, sigma, nu, tau = self.theta_to_params(theta)
        q1 = mu + (sigma / nu) * st.t.ppf((q * (1 + nu**2)) * 0.5, df=tau)
        q2 = mu + sigma * nu * st.t.ppf(
            (q * (1 + nu**2) - 1 + nu**2) / (2 * nu**2), df=tau
        )
        quantile = np.where(q <= (1 + nu**2) ** -1, q1, q2)
        return quantile

    def pdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        mu, sigma, nu, tau = self.theta_to_params(theta)
        z = (y - mu) / sigma
        density1 = (2 * nu / (sigma * (1 + nu**2))) * (
            1 + ((nu**2) * (z**2)) / tau
        ) ** (-(tau + 1) / 2)
        density2 = (2 * nu / (sigma * (1 + nu**2))) * (
            1 + (z**2) / ((nu**2) * tau)
        ) ** (-(tau + 1) / 2)
        density = np.where(y < mu, density1, density2)
        return density

    def rvs(self, size: Tuple[int, ...], theta: np.ndarray) -> np.ndarray:
        u = st.uniform.rvs(size=size)
        samples = self.ppf(u, theta)
        return samples

    def logcdf(self, y, theta):
        return np.log(self.cdf(y, theta))

    def logpdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        return np.log(self.pdf(y, theta))

    @staticmethod
    def _get_terms_deriv(mu, sigma, nu, tau, y):
        s1 = sigma / nu
        s2 = sigma * nu
        dsq1 = ((y - mu) ** 2) / (s1**2)
        dsq2 = ((y - mu) ** 2) / (s2**2)
        w1 = np.where(tau < 1000000, (tau + 1) / (tau + dsq1), 1)
        w2 = np.where(tau < 1000000, (tau + 1) / (tau + dsq2), 1)
        return s1, s2, dsq1, dsq2, w1, w2

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu, tau = self.theta_to_params(theta)
        s1, s2, dsq1, dsq2, w1, w2 = self._get_terms_deriv(mu, sigma, nu, tau, y)

        match param:
            case 0:
                deriv = np.where(
                    y < mu,
                    (w1 * (y - mu)) / s1**2,
                    (w2 * (y - mu)) / s2**2,
                )
                return deriv
            case 1:
                deriv = np.where(
                    y < mu,
                    (w1 * dsq1 - 1) / sigma,
                    (w2 * dsq2 - 1) / sigma,
                )
                return deriv
            case 2:
                deriv = np.where(
                    y < mu,
                    -(w1 * dsq1 - 1) / nu,
                    (w2 * dsq2 + 1) / nu,
                ) - 2 * nu / (1 + nu**2)
                return deriv
            case 3:
                deriv_a = -0.5 * np.log(1 + dsq1 / tau) + (w1 * dsq1 - 1) / (2 * tau)
                deriv_b = -0.5 * np.log(1 + dsq2 / tau) + (w2 * dsq2 - 1) / (2 * tau)
                deriv = np.where(y < mu, deriv_a, deriv_b)
                deriv = (
                    deriv
                    + 0.5 * spc.digamma((tau + 1) / 2)
                    - 0.5 * spc.digamma(tau / 2)
                )
                return deriv

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        deriv = -(self.dl1_dp1(y, theta, param) ** 2)
        return np.clip(deriv, -np.inf, -1e-15)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        return -self.dl1_dp1(y, theta, params[0]) * self.dl1_dp1(y, theta, params[1])

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y), np.std(y, ddof=1), 1, 10]
        return np.tile(initial_params, (y.shape[0], 1))

    def mean(self, theta: np.ndarray) -> np.ndarray:
        mu, sigma, nu, tau = self.theta_to_params(theta)
        c = 2 * nu * ((1 + nu**2) * spc.beta(0.5, tau / 2) * tau**0.5) ** -1
        mean = mu + sigma * c
        return mean

    def variance(self, theta: np.ndarray) -> np.ndarray:
        _, sigma, nu, tau = self.theta_to_params(theta)
        m = ((2 * tau**0.5) * (nu - nu**-1)) / ((tau - 1) * spc.beta(0.5, 0.5 * tau))
        variance = sigma**2 * ((tau / (tau - 2)) * (nu**2 + nu**-2 - 1) - m**2)
        return variance

    def calculate_conditional_initial_values(self, y, theta, param):
        return super().calculate_conditional_initial_values(y, theta, param)

    def logpmf(self, y, theta):
        raise NotImplementedError("LogPMF is not implemented for SkewT distribution.")

    def pmf(self, y, theta):
        raise NotImplementedError("PMF is not implemented for SkewT distribution.")


class SkewTMeanStd(ScipyMixin, Distribution):
    corresponding_gamlss: str = "SST"
    parameter_names = {0: "mu", 1: "sigma", 2: "nu", 3: "tau"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (np.nextafter(0, 1), np.inf),
        3: (np.nextafter(0, 1), np.inf),
    }
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
        2: ParameterShapes.SCALAR,
        3: ParameterShapes.SCALAR,
    }
    distribution_support = (-np.inf, np.inf)
    scipy_dist = None
    scipy_names = None

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
        skew_link: LinkFunction = Log(),
        shape_link: LinkFunction = LogShiftTwo(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
                2: skew_link,
                3: shape_link,
            }
        )
        self._st3dist = SkewT()

    def _map(self, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        c = 2 * nu * ((1 + nu**2) * spc.beta(0.5, tau / 2) * tau**0.5) ** -1
        m = ((2 * tau**0.5) * (nu - nu**-1)) / ((tau - 1) * spc.beta(0.5, 0.5 * tau))
        s2 = (tau / (tau - 2)) * (nu**2 + nu**-2 - 1) - m**2
        mu_0 = mu - (sigma * m / np.sqrt(s2))
        sigma_0 = sigma / np.sqrt(s2)
        return mu_0, sigma_0, nu, tau, c

    def cdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        _, _, nu, tau = self.theta_to_params(theta)
        mu_0, sigma_0, nu, tau, c = self._map(theta)

        prob = np.where(
            y < mu_0,
            (2 / (1 + nu**2)) * st.t.cdf(x=(nu * (y - mu_0) / sigma_0), df=tau),
            (1 / (1 + nu**2))
            * (1 + 2 * nu**2 * (st.t.cdf(x=(y - mu_0) / (sigma_0 * nu), df=tau) - 0.5)),
        )
        return prob

    def ppf(self, q: np.ndarray, theta: np.ndarray) -> np.ndarray:
        _, _, nu, tau = self.theta_to_params(theta)
        mu_0, sigma_0, nu, tau, c = self._map(theta)

        quantile = np.where(
            q <= (1 + nu**2) ** -1,
            mu_0 + (sigma_0 / nu) * st.t.ppf((q * (1 + nu**2)) * 0.5, tau),
            mu_0
            + sigma_0 * nu * st.t.ppf((q * (1 + nu**2) - 1 + nu**2) / (2 * nu**2), tau),
        )
        return quantile

    def pdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        _, _, nu, tau = self.theta_to_params(theta)
        mu_0, sigma_0, nu, tau, c = self._map(theta)
        z = (y - mu_0) / sigma_0
        density = np.where(
            y < mu_0,
            (c / sigma_0) * (1 + ((nu**2) * (z**2)) / tau) ** (-(tau + 1) / 2),
            (c / sigma_0) * (1 + (z**2) / ((nu**2) * tau)) ** (-(tau + 1) / 2),
        )
        return density

    def rvs(self, size: Tuple[int, ...], theta: np.ndarray) -> np.ndarray:
        u = st.uniform.rvs(size=size)
        samples = self.ppf(u, theta)
        return samples

    def logcdf(self, y, theta):
        return np.log(self.cdf(y, theta))

    def logpdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        return np.log(self.pdf(y, theta))

    def logpmf(self, y, theta):
        raise NotImplementedError(
            "LogPMF is not implemented for SkewTMeanStd distribution."
        )

    def pmf(self, y, theta):
        raise NotImplementedError(
            "PMF is not implemented for SkewTMeanStd distribution."
        )

    def _get_terms_deriv(self, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        m1 = (2 * tau ** (1 / 2) * (nu**2 - 1)) / (
            (tau - 1) * spc.beta(1 / 2, tau / 2) * nu
        )
        m2 = (tau * (nu**3 + (1 / nu**3))) / ((tau - 2) * (nu + (1 / nu)))
        s1 = np.sqrt(m2 - m1**2)
        mu1 = mu - ((sigma * m1) / s1)
        sigma1 = sigma / s1
        return m1, m2, s1, mu1, sigma1

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y), np.std(y, ddof=1), 1, 10]
        return np.tile(initial_params, (y.shape[0], 1))

    def dl1_dp1(self, y, theta, param):
        self._validate_dln_dpn_inputs(y, theta, param)
        _, sigma, nu, tau = self.theta_to_params(theta)
        m1, m2, s1, mu1, sigma1 = self._get_terms_deriv(theta)

        st3theta = np.column_stack((mu1, sigma1, nu, tau))

        match param:
            case 0:
                deriv = self._st3dist.dl1_dp1(y, st3theta, param)
                return deriv
            case 1:
                return -(m1 / s1) * self._st3dist.dl1_dp1(y, st3theta, 0) + (
                    1 / s1
                ) * self._st3dist.dl1_dp1(y, st3theta, 1)
            case 2:
                dl1dmu1 = self._st3dist.dl1_dp1(y, st3theta, 0)
                dl1dd1 = self._st3dist.dl1_dp1(y, st3theta, 1)
                dl1dv = self._st3dist.dl1_dp1(y, st3theta, 2)
                dmu1dm1 = -sigma / s1
                dmu1ds1 = (sigma * m1) / (s1**2)
                dd1ds1 = -sigma / (s1**2)
                dm1dv = (
                    (2 * tau ** (1 / 2)) / ((tau - 1) * spc.beta(1 / 2, tau / 2))
                ) * ((nu**2 + 1) / (nu**2))
                dm2dv = m2 * (
                    (6 * nu**5 / (nu**6 + 1)) - (2 / nu) - (2 * nu / (nu**2 + 1))
                )
                ds1dv = (dm2dv - 2 * m1 * dm1dv) / (2 * s1)
                dldv = (
                    dl1dmu1 * dmu1dm1 * dm1dv
                    + dl1dmu1 * dmu1ds1 * ds1dv
                    + dl1dd1 * dd1ds1 * ds1dv
                    + dl1dv
                )
                return dldv
            case 3:
                dl1dmu1 = self._st3dist.dl1_dp1(y, st3theta, 0)
                dl1dd1 = self._st3dist.dl1_dp1(y, st3theta, 1)
                dl1dt = self._st3dist.dl1_dp1(y, st3theta, 3)
                dmu1dm1 = -sigma / s1
                dmu1ds1 = (sigma * m1) / (s1**2)
                dd1ds1 = -sigma / (s1**2)
                dm1dt = m1 * (
                    (1 / (2 * tau))
                    - (1 / (tau - 1))
                    - 0.5 * (spc.digamma(tau / 2))
                    + 0.5 * (spc.digamma((tau + 1) / 2))
                )
                dm2dt = -m2 * (2 / (tau * (tau - 2)))
                ds1dt = (dm2dt - 2 * m1 * dm1dt) / (2 * s1)
                dldt = (
                    dl1dmu1 * dmu1dm1 * dm1dt
                    + dl1dmu1 * dmu1ds1 * ds1dt
                    + dl1dd1 * dd1ds1 * ds1dt
                    + dl1dt
                )
                return dldt

    def dl2_dp2(self, y, theta, param):
        self._validate_dln_dpn_inputs(y, theta, param)
        deriv = -(self.dl1_dp1(y, theta, param) ** 2)
        return np.clip(deriv, -np.inf, -1e-15)

    def dl2_dpp(self, y, theta, params):
        self._validate_dl2_dpp_inputs(y, theta, params)
        return -self.dl1_dp1(y, theta, params[0]) * self.dl1_dp1(y, theta, params[1])



# Author Simon Hirsch
# MIT Licence
from typing import Dict

import numba as nb
import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, MultivariateDistributionMixin
from ..links import Identity, Log, MatrixDiagTriu
from ..types import ParameterShapes


class MultivariateNormalInverseCholesky(MultivariateDistributionMixin, Distribution):
    # The cholesky decomposition of
    # COV = L @ L.T
    # PRC = (L^-1).T @ (L^-1)
    # The cross derivatives are defined on (L^-1).T
    # The transpose is important!!

    """
    The multivariate normal (Gaussian) distribution parameterized by the inverse Cholesky factor of the precision (inverse scale) matrix.

    The PDF of the multivariate normal distribution is given by:
    $$
    p(y \\mid \\mu, L) =
    |L| \\cdot (2\\pi)^{-k/2}
    \\exp\\left(-\\frac{1}{2} (y - \\mu)^T (L L^T) (y - \\mu)\\right)
    $$

    where \\( k \\) is the dimensionality of the data, \\( \\mu \\) is the location parameter, and \\( L \\) is the inverse Cholesky factor of the precision matrix (so the precision is \\( L L^T \\)).
    """

    corresponding_gamlss: str = None
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {0: (-np.inf, np.inf), 1: (-np.inf, np.inf)}
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.multivariate_normal
    scipy_names = {"mu": "mean", "sigma": "cov"}
    parameter_shape = {
        0: ParameterShapes.VECTOR,
        1: ParameterShapes.UPPER_TRIANGULAR_MATRIX,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = MatrixDiagTriu(Log(), Identity()),
    ):
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )
        self._regularization_allowed = {0: False, 1: True}

    @staticmethod
    def fitted_elements(dim: int):
        return {0: dim, 1: int(dim * (dim + 1) // 2)}

    @staticmethod
    def index_flat_to_cube(k: int, d: int, param: int):
        if param == 0:
            return k
        if param == 1:
            i, j = np.triu_indices(d)
            return i[k], j[k]

    def get_adr_regularization_distance(self, dim: int, param: int) -> float:
        if param == 0:
            return None
        if param == 1:
            i, j = np.triu_indices(dim, k=0)
            return np.abs(i - j)

    def get_regularization_size(self, dim: int) -> int:
        return dim

    @staticmethod
    def set_theta_element(theta: Dict, value: np.ndarray, param: int, k: int) -> Dict:
        """Sets an element of theta for parameter param and place k.

        !!! Note
            This will mutate `theta`!

        Args:
            theta (Dict): Current fitted $\theta$
            value (np.ndarray): Value to set
            param (int): Distribution parameter
            k (int): Flat element index $k$

        Returns:
            Dict: Theta where element (param, k) is set to value.
        """
        if param == 0:
            theta[param][:, k] = value
        if param == 1:
            d = theta[0].shape[1]
            i, j = np.triu_indices(d)
            theta[param][:, i[k], j[k]] = value
        return theta

    def theta_to_params(self, theta):
        fitted_loc = theta[0]
        fitted_inv_tr_chol = theta[1]
        return fitted_loc, fitted_inv_tr_chol

    def set_initial_guess(
        self,
        y: np.ndarray,
        theta: Dict[int, np.ndarray],
        param: int,
    ) -> Dict[int, np.ndarray]:
        if param == 0:
            return theta
        if param == 1:
            residual = y - theta[0]
            cov = np.cov(residual, rowvar=False)
            chol = np.linalg.inv(np.linalg.cholesky(cov)).T
            theta[1] = np.tile(chol, (y.shape[0], 1, 1))
            return theta

    def dl1_dp1(self, y: np.ndarray, theta: Dict, param: int = 0):
        """Return the first derivatives wrt to the parameter.

        !!! Note
            We expect the fitted L^-1)^T to be handed in matrix/cube form, i.e of shape n x d x d.
            But we return the derivatives in flat format.

        Args:
            y (np.ndarray): Y values of shape n x d
            theta (Dict): Dict with {0 : fitted mu, 1 : fitted (L^-1)^T}
            param (int, optional): Which parameter derivatives to return. Defaults to 0.

        Returns:
            derivative (np.ndarray): The 1st derivatives.
        """
        fitted_loc, fitted_inv_tr_chol = self.theta_to_params(theta)

        if param == 0:
            deriv = _derivative_1st_mu(
                y=y, fitted_loc=fitted_loc, fitted_inv_tr_chol=fitted_inv_tr_chol
            )
        if param == 1:
            deriv = _derivative_1st_inv_chol(
                y=y, fitted_loc=fitted_loc, fitted_inv_tr_chol=fitted_inv_tr_chol
            )

        return deriv

    def dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, clip=False
    ) -> np.ndarray:
        """Return the second derivatives wrt to the parameter.

        !!! Note
            We expect the fitted L^-1)^T to be handed in matrix/cube form, i.e of shape n x d x d.
            But we return the derivatives in flat format.

        Args:
            y (np.ndarray): Y values of shape n x d
            theta (Dict): Dict with {0 : fitted mu, 1 : fitted (L^-1)^T}
            param (int, optional): Which parameter derivatives to return. Defaults to 0.

        Returns:
            derivative (np.ndarray): The 2nd derivatives.
        """
        fitted_loc, fitted_inv_tr_chol = self.theta_to_params(theta)

        if param == 0:
            deriv = _derivative_2nd_mu(
                y=y, fitted_loc=fitted_loc, fitted_inv_tr_chol=fitted_inv_tr_chol
            )
        if param == 1:
            deriv = _derivative_2nd_inv_chol(
                y=y, fitted_loc=fitted_loc, fitted_inv_tr_chol=fitted_inv_tr_chol
            )

        return deriv

    def element_score(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl1_dp1(y=y, theta=theta, param=param, k=k)

    def element_hessian(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl2_dp2(y=y, theta=theta, param=param, k=k)

    def element_dl1_dp1(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip=False
    ):
        fitted_loc, fitted_inv_tr_chol = self.theta_to_params(theta)
        d = y.shape[1]
        if param == 0:
            deriv = _derivative_1st_mu(
                y=y, fitted_loc=fitted_loc, fitted_inv_tr_chol=fitted_inv_tr_chol
            )
            deriv = deriv[:, k]
        if param == 1:
            ii, jj = np.triu_indices(d)
            if ii[k] == jj[k]:
                deriv = _derivative_1st_inv_chol_diag(
                    y, fitted_loc, fitted_inv_tr_chol, ii[k]
                )
            else:
                deriv = _derivative_1st_inv_chol_triu(
                    y, fitted_loc, fitted_inv_tr_chol, ii[k], jj[k]
                )
        return deriv

    def element_dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip=False
    ):
        fitted_loc, fitted_inv_tr_chol = self.theta_to_params(theta)
        d = y.shape[1]
        if param == 0:
            deriv = _derivative_2nd_mu(
                y=y, fitted_loc=fitted_loc, fitted_inv_tr_chol=fitted_inv_tr_chol
            )
            deriv = deriv[:, k]
        if param == 1:
            ii, jj = np.triu_indices(d)
            if ii[k] == jj[k]:
                deriv = _derivative_2nd_inv_chol_diag(
                    y, fitted_loc, fitted_inv_tr_chol, ii[k]
                )
            else:
                deriv = _derivative_2nd_inv_chol_triu(
                    y, fitted_loc, fitted_inv_tr_chol, ii[k], jj[k]
                )
        return deriv

    def dl2_dpp(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented.")

    def element_link_function(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].link(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_link(y, i=i[k], j=j[k])

    def element_link_function_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].link_derivative(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_derivative(y, i=i[k], j=j[k])

    def element_link_function_second_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].link_second_derivative(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_link_second_derivative(y, i=i[k], j=j[k])

    def element_link_inverse(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].inverse(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_inverse(y, i=i[k], j=j[k])

    def element_link_inverse_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].inverse_derivative(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_inverse_derivative(y, i=i[k], j=j[k])

    def initial_values(self, y, param=0):
        M = y.shape[0]
        if param == 0:
            return np.tile(np.mean(y, axis=0), (M, 1))
        if param == 1:
            chol = np.linalg.inv(np.linalg.cholesky(np.cov(y, rowvar=False))).T
            return np.tile(chol, (M, 1, 1))

    def cube_to_flat(self, x: np.ndarray, param: int):
        if param == 0:
            return x
        if param == 1:
            d = x.shape[1]
            i = np.triu_indices(d)
            out = x[:, i[0], i[1]]
            return out

    def flat_to_cube(self, x: np.ndarray, param: int):
        if param == 0:
            return x
        if param == 1:
            n, k = x.shape
            # The following conversion holds for upper diagonal matrices
            # We INCLUDE the diagonal!!
            # (D + 1) * D // 2 = k
            # (D + 1) * D = 2 * k
            # D^2 + D = 2 * k
            # ... Wolfram Alpha ...
            # D = 0.5 * (sqrt(8k + 1) - 1)
            d = int(1 / 2 * (np.sqrt(8 * k + 1) - 1))
            i = np.triu_indices(d)
            out = np.zeros((n, d, d))
            out[:, i[0], i[1]] = x
            return out

    def param_conditional_likelihood(
        self, y: np.ndarray, theta: Dict, eta: np.ndarray, param: int
    ) -> np.ndarray:
        """Calulate the log-likelihood for (flat) eta for parameter (param)
        and theta for all other parameters.

        Args:
            y (np.ndarray): True values
            theta (Dict): Fitted theta.
            eta (np.ndarray): Fitted eta.
            param (int): Param for which we take eta.

        Returns:
            np.ndarray: Log-likelihood.
        """
        fitted = self.flat_to_cube(eta, param=param)
        fitted = self.link_inverse(fitted, param=param)
        # fitted_theta = {**theta, param: fitted_eta}
        return self.log_likelihood(y, theta={**theta, param: fitted})

    def theta_to_scipy(self, theta: Dict[int, np.ndarray]):
        out = {
            "mean": theta[0],
            "cov": np.linalg.inv(theta[1] @ theta[1].swapaxes(-1, -2)),
        }
        return out

    def cdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def ppf(self, q, theta):
        raise NotImplementedError("Not implemented")

    def rvs(self, size, theta):
        raise NotImplementedError("Not implemented")

    def logcdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def logpdf(self, y, theta):
        return _log_likelihood(y, theta[0], theta[1])

    def logpmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def calculate_conditional_initial_values(
        self, y: np.ndarray, theta: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        raise NotImplementedError("Not implemented")


##########################################################
### numba JIT-compiled functions for the derivatives #####
##########################################################


@nb.jit(nopython=True)
def _log_likelihood(y, mu, mod_chol_transpose):
    M = y.shape[0]
    k = y.shape[1]
    y_centered = y - mu
    ll = np.empty(M)
    for m in range(M):
        ll[m] = float(
            np.log(np.linalg.det(mod_chol_transpose[m]))
            - 0.5
            * (
                y_centered[m]
                @ mod_chol_transpose[m]
                @ mod_chol_transpose[m].T
                @ y_centered[m]
            )
        )
    ll -= k / 2 * np.log(np.pi * 2)
    return ll


def _derivative_1st_mu(
    y: np.ndarray, fitted_loc: np.ndarray, fitted_inv_tr_chol: np.ndarray
) -> np.ndarray:
    y_centered = y - fitted_loc
    precision = fitted_inv_tr_chol @ fitted_inv_tr_chol.transpose(0, 2, 1)
    return np.sum((y_centered[:, :, None] * precision), -1)


@nb.jit()
def _derivative_1st_inv_chol_diag(y, fitted_loc, fitted_inv_tr_chol, i):
    y_centered = y - fitted_loc
    sum_m = np.zeros(y.shape[0])
    for m in range(i + 1):
        sum_m += y_centered[:, m] * fitted_inv_tr_chol[:, m, i]
    deriv = 1 / fitted_inv_tr_chol[:, i, i] - y_centered[:, i] * sum_m
    return deriv


@nb.jit()
def _derivative_1st_inv_chol_triu(y, fitted_loc, fitted_inv_tr_chol, i, j):
    y_centered = y - fitted_loc
    sum_m = np.zeros(y.shape[0])
    for m in range(j + 1):
        sum_m += y_centered[:, m] * fitted_inv_tr_chol[:, m, j]
    deriv = -(y_centered[:, i] * sum_m)
    return deriv


@nb.jit()
def _derivative_1st_inv_chol(y, fitted_loc, fitted_inv_tr_chol):
    n = y.shape[0]
    d = y.shape[1]
    dim = ((d + 1) * d) // 2
    k = 0
    deriv = np.zeros((n, dim))
    for i in range(d):
        for j in range(i, d):
            if i == j:
                deriv[:, k] = _derivative_1st_inv_chol_diag(
                    y, fitted_loc, fitted_inv_tr_chol, i
                )
            else:
                deriv[:, k] = _derivative_1st_inv_chol_triu(
                    y, fitted_loc, fitted_inv_tr_chol, i, j
                )
            k += 1
    return deriv


@nb.jit()
def _derivative_2nd_mu(y, fitted_loc, fitted_inv_tr_chol):
    d = y.shape[1]
    deriv = np.zeros_like(y)
    for i in range(d):
        deriv[:, i] = -np.sum(fitted_inv_tr_chol[:, i, i:] ** 2, -1)
    return deriv


@nb.jit()
def _derivative_2nd_inv_chol_diag(y, fitted_loc, fitted_inv_tr_chol, i):
    y_centered = y - fitted_loc
    deriv = -1 / fitted_inv_tr_chol[:, i, i] ** 2 - y_centered[:, i] ** 2
    return deriv


# @nb.jit()
# def _derivative_2nd_inv_chol_diag(y, fitted_loc, fitted_inv_tr_chol, i):
#     y_centered = y - fitted_loc
#     sum_m = np.zeros(y.shape[0])
#     for m in range(i):
#         sum_m += y_centered[:, m] * fitted_inv_tr_chol[:, m, i]
#     deriv = (
#         -2 * fitted_inv_tr_chol[:, i, i] ** 2 * y_centered[:, i] ** 2
#         - fitted_inv_tr_chol[:, i, i] * y_centered[:, i] * sum_m
#     )
#     return deriv


@nb.jit()
def _derivative_2nd_inv_chol_triu(y, fitted_loc, fitted_inv_tr_chol, i, j):
    y_centered = y - fitted_loc
    return -(y_centered[:, i] ** 2)


@nb.jit()
def _derivative_2nd_inv_chol(y, fitted_loc, fitted_inv_tr_chol):
    n = y.shape[0]
    d = y.shape[1]
    dim = ((d + 1) * d) // 2
    k = 0
    deriv = np.zeros((n, dim))
    for i in range(d):
        for j in range(i, d):
            if i == j:
                deriv[:, k] = _derivative_2nd_inv_chol_diag(
                    y, fitted_loc, fitted_inv_tr_chol, i
                )
            else:
                deriv[:, k] = _derivative_2nd_inv_chol_triu(
                    y, fitted_loc, fitted_inv_tr_chol, i, j
                )
            k += 1
    return deriv



from enum import Enum


class ParameterShapes(Enum):
    SCALAR = "A scalar parameter N x 1"
    VECTOR = "A vector parameter N x D"
    MATRIX = "A matrix parameter N x D x M"
    SQUARE_MATRIX = "A matrix parameter N x D x D"
    DIAGONAL_MATRIX = "A diagonal matrix N x D x D."
    LOWER_TRIANGULAR_MATRIX = "A lower triangular matrix N x D x D."
    UPPER_TRIANGULAR_MATRIX = "A upper triangular matrix N x D x D."



from abc import ABC, abstractmethod
from typing import Tuple


class DiagnosticDisplay(ABC):
    @classmethod
    @abstractmethod
    def from_estimator(
        cls, estimator, X, y, ax=None, figsize: Tuple[float, float] = (10, 5), **kwargs
    ):
        raise NotImplementedError("Should implement from_estimator method")

    @abstractmethod
    def plot(self, ax=None, figsize: Tuple[float, float] = (10, 5), **kwargs):
        raise NotImplementedError("Should implement plot method")



import numpy as np


def add_intercept(X: np.ndarray):
    return np.hstack((np.ones((X.shape[0], 1)), X))


def make_intercept(n_observations: int) -> np.ndarray:
    """Make the intercept series as N x 1 array.

    Args:
        y (np.ndarray): Response variable $Y$

    Returns:
        np.ndarray: Intercept array.
    """
    return np.ones((n_observations, 1))



from itertools import product
from typing import Dict

import numpy as np
import scipy.special as sp
import scipy.stats as st

from ..base import Distribution, LinkFunction, MultivariateDistributionMixin
from ..links import Identity, Log, LogShiftTwo, MatrixDiag
from ..types import ParameterShapes


class MultivariateStudentTInverseLowRank(MultivariateDistributionMixin, Distribution):
    """
    The multivariate \\( t \\)-distribution using a low-rank approximation (LRA) of the precision (inverse scale) matrix.

    The PDF of the multivariate \\( t \\)-distribution is given by:
    $$
    p(y \mid \\mu, D, V, \\nu) =
    \\frac{\\Gamma\\left(\\frac{\\nu + k}{2}\\right)}
         {\\Gamma\\left(\\frac{\\nu}{2}\\right) \\left(\\pi \\nu\\right)^{k/2}}
    \\cdot \\frac{1}{\\sqrt{\\det(D + V V^T)}}
    \\left(1 + \\frac{1}{\\nu} (y - \\mu)^T (D + V V^T)^{-1} (y - \\mu)\\right)^{-\\frac{\\nu + k}{2}}
    $$

    where \\( k \\) is the dimensionality of the data, \\( \\mu \\) is the location parameter, \\( D \\) is a diagonal matrix, \\( V \\) is a low-rank matrix, and \\( \\nu \\) is the degrees of freedom.
    """

    corresponding_gamlss: str = None
    parameter_names = {0: "mu", 1: "D", 2: "V", 3: "nu"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (0, np.inf),
        2: (-np.inf, np.inf),
        3: (0, np.inf),
    }
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.multivariate_t
    scipy_names = {"mu": "loc", "sigma": "shape", "nu": "dof"}
    parameter_shape = {
        0: ParameterShapes.VECTOR,
        1: ParameterShapes.DIAGONAL_MATRIX,
        2: ParameterShapes.MATRIX,
        3: ParameterShapes.SCALAR,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link_1: LinkFunction = MatrixDiag(Log()),
        scale_link_2: LinkFunction = Identity(),
        tail_link: LinkFunction = LogShiftTwo(),
        rank: int = 3,
        dof_guesstimate: float = 10.0,
    ):
        super().__init__(
            links={0: loc_link, 1: scale_link_1, 2: scale_link_2, 3: tail_link}
        )
        self.rank = rank
        self.dof_guesstimate = dof_guesstimate
        self.dof_independence = 1e6
        self._regularization_allowed = {0: False, 1: False, 2: True, 3: False}

    def fitted_elements(self, dim: int):
        return {0: dim, 1: dim, 2: dim * self.rank, 3: 1}

    @property
    def param_structure(self):
        return self._param_structure

    def index_flat_to_cube(self, k: int, d: int, param: int):
        if (param == 0) | (param == 3):
            return k
        if param == 1:
            i, j = np.diag_indices(d)
            return i[k], j[k]
        if param == 2:
            idx = [(j, i) for i, j in product(range(self.rank), range(d))]
            return idx[k][0], idx[k][1]

    def get_adr_regularization_distance(self, dim: int, param: int) -> np.ndarray:
        if param in (0, 1, 3):
            return None
        if param == 2:
            return np.concatenate([np.repeat(i + 1, dim) for i in range(self.rank)])

    def get_regularization_size(self, dim: int) -> int:
        return self.rank + 1

    def set_theta_element(
        self, theta: Dict, value: np.ndarray, param: int, k: int
    ) -> Dict:
        """Sets an element of theta for parameter param and place k.

        !!! Note
            This will mutate `theta`!

        Args:
            theta (Dict): Current fitted $\theta$
            value (np.ndarray): Value to set
            param (int): Distribution parameter
            k (int): Flat element index $k$

        Returns:
            Dict: Theta where element (param, k) is set to value.
        """
        if (param == 0) | (param == 3):
            theta[param][:, k] = value
        if (param == 1) | (param == 2):
            d = theta[0].shape[1]
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            theta[param][:, i, j] = value
        return theta

    def theta_to_params(self, theta):
        loc = theta[0]
        mat_d = theta[1]
        mat_v = theta[2]
        mat_dof = theta[3].squeeze()
        return loc, mat_d, mat_v, mat_dof

    def dl1_dp1(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dp2(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dpp(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented.")

    def element_score(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl1_dp1(y=y, theta=theta, param=param, k=k)

    def element_hessian(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl2_dp2(y=y, theta=theta, param=param, k=k)

    def element_dl1_dp1(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip: bool = False
    ):
        mu, mat_d, mat_v, dof = self.theta_to_params(theta)
        d = y.shape[1]
        if param == 0:
            deriv = mv_t_lr_partial_1_mu(y, mu, mat_d, mat_v, dof, i=k)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = mv_t_lr_partial_1_D_element(y, mu, mat_d, mat_v, dof, i=i)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = mv_t_lr_partial_1_V_element(y, mu, mat_d, mat_v, dof, i=i, j=j)
        if param == 3:
            deriv = mv_t_lr_partial_1_dof(y, mu, mat_d, mat_v, dof)
        return deriv

    def element_dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip: bool = True
    ):
        mu, mat_d, mat_v, dof = self.theta_to_params(theta)
        d = y.shape[1]
        if param == 0:
            deriv = mv_t_lr_partial_2_mu(y, mu, mat_d, mat_v, dof, i=k)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = mv_t_lr_partial_2_D_element(y, mu, mat_d, mat_v, dof, i=i)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = mv_t_lr_partial_2_V_element(y, mu, mat_d, mat_v, dof, i=i, j=j)
        if param == 3:
            deriv = mv_t_lr_partial_2_dof(y, mu, mat_d, mat_v, dof)

        if clip:
            deriv = np.clip(deriv, -np.inf, -1e-10)

        return deriv

    def element_link_function(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_link(y, i=i, j=j)

    def element_link_inverse(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2) | (param == 3):
            return self.links[param].inverse(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_inverse(y, i=i, j=j)

    def element_link_function_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2) | (param == 3):
            return self.links[param].link_derivative(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_link_derivative(y, i=i, j=j)

    def element_link_function_second_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2) | (param == 3):
            return self.links[param].link_second_derivative(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_link_second_derivative(y, i=i, j=j)

    def element_link_inverse_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2) | (param == 3):
            return self.links[param].inverse_derivative(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_inverse_derivative(y, i=i, j=j)

    def initial_values(self, y: np.ndarray, param: int = 0):
        M = y.shape[0]
        if param == 0:
            return np.tile(np.mean(y, axis=0), (M, 1))
        if param == 1:
            mat_d = np.diag(
                1 / (np.var(y, 0) * (self.dof_guesstimate - 2) / self.dof_guesstimate)
            )
            return np.tile(mat_d, (M, 1, 1))
        if param == 2:
            omega = np.linalg.inv(np.cov(y, rowvar=False))
            mat_d = np.diag(
                1 / (np.var(y, 0) * (self.dof_guesstimate - 2) / self.dof_guesstimate)
            )
            eig = np.linalg.eig(omega - mat_d)
            largest_ev = np.argsort(eig.eigenvalues)[-self.rank :][::-1]
            mat_v = eig.eigenvectors[:, largest_ev]
            return np.tile(mat_v, (M, 1, 1))
        if param == 3:
            return np.full((M, 1), self.dof_independence)

    def set_initial_guess(
        self,
        y: np.ndarray,
        theta: Dict[int, np.ndarray],
        param: int,
    ) -> Dict[int, np.ndarray]:
        if param in (0, 2):
            return theta
        if param == 1:
            residuals = y - theta[0]
            variance = np.var(residuals, 0)
            mat_d = np.diag(
                1 / (variance * (self.dof_guesstimate - 2) / self.dof_guesstimate)
            )
            theta[1] = np.tile(mat_d, (y.shape[0], 1, 1))
            return theta
        if param == 3:
            theta[3] = np.full_like(theta[3], self.dof_guesstimate)
            return theta

    def cube_to_flat(self, x: np.ndarray, param: int):
        if (param == 0) | (param == 3):
            return x
        if param == 1:
            return np.copy(x.diagonal(axis1=1, axis2=2))
        if param == 2:
            return x.swapaxes(-1, -2).reshape((x.shape[0], np.prod(x.shape[1:])))

    def flat_to_cube(self, x: np.ndarray, param: int):
        if (param == 0) | (param == 3):
            return x
        if param == 1:
            d = x.shape[1]
            out = np.zeros((x.shape[0], d, d))
            out[:, np.arange(d), np.arange(d)] = x
            return out
        if param == 2:
            d = int(x.shape[1] // self.rank)
            return x.reshape((x.shape[0], self.rank, d)).transpose(0, 2, 1)

    def theta_to_scipy(self, theta: Dict[int, np.ndarray]):
        out = {
            "loc": theta[0],
            "shape": np.linalg.inv(theta[1] + theta[2] @ theta[2].swapaxes(-1, -2)),
            "df": theta[3],
        }
        return out

    def cdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def ppf(self, q, theta):
        raise NotImplementedError("Not implemented")

    def rvs(self, size, theta):
        raise NotImplementedError("Not implemented")

    def logcdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def logpdf(self, y, theta):
        loc, mat_d, mat_v, dof = self.theta_to_params(theta)
        return batched_log_lilkelihood_t_precision_low_rank_fast(
            y, loc, mat_d=mat_d, mat_v=mat_v, dof=dof
        )

    def logpmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def calculate_conditional_initial_values(
        self, y: np.ndarray, theta: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        raise NotImplementedError("Not implemented")


def batched_log_lilkelihood_t_precision_low_rank_fast(y, mu, mat_d, mat_v, dof):
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    A = np.squeeze(sp.gammaln((dof + k) / 2))
    B = np.squeeze(sp.gammaln(dof / 2))
    C = 1 / 2 * k * np.squeeze(np.log(np.pi * dof))
    Z = np.sum((y - mu) * (precision @ (y - mu)[..., None]).squeeze(), 1)
    part1 = A - B - C
    part2 = 1 / 2 * np.log(np.linalg.det(precision))
    part3 = np.squeeze((dof + k) / 2) * np.log((1 + np.squeeze((1 / dof)) * Z))
    return part1 + part2 - part3


def mv_t_lr_partial_1_mu(y, mat_mu, mat_d, mat_v, dof, i):
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)
    part1 = (k + dof) / (2 * (Z + dof))
    part2 = 2 * np.sum(precision[:, i, :] * (y - mat_mu), axis=1)
    return part1 * part2


def mv_t_lr_partial_2_mu(y, mat_mu, mat_d, mat_v, dof, i):
    k = y.shape[1]

    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)
    deriv_1 = 2 * np.sum(precision[:, i, :] * (y - mat_mu), axis=1)
    deriv_2 = 2 * precision[:, i, i]

    part1 = k + dof
    part2 = (Z + dof) * deriv_2 - deriv_1**2
    part3 = 2 * (Z + dof) ** 2

    return -(part1 * part2) / part3


def mv_t_lr_partial_1_D_element(y, mat_mu, mat_d, mat_v, dof, i):
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)

    # Derivative of the log(det(D + VV^T))
    part_1 = 0.5 * np.linalg.inv(precision)[:, i, i]

    # Derivative of the last term
    part2 = np.squeeze(k + dof) / (2 * (Z + dof.squeeze()))
    part3 = (y[:, i] - mat_mu[:, i]) ** 2
    return part_1 - part2 * part3


def mv_t_lr_partial_2_D_element(y, mat_mu, mat_d, mat_v, dof, i):
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    cov = np.linalg.inv(precision)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)

    # Derivative for the log(det())
    part1 = -0.5 * cov[:, i, i] ** 2

    # Derivative for the last term
    deriv_lambda_1 = (y[:, i] - mat_mu[:, i]) ** 2
    deriv_lambda_2 = 0

    part2 = k + dof.squeeze()
    part3 = (Z + dof.squeeze()) * deriv_lambda_2 - deriv_lambda_1**2
    part4 = 2 * (Z + dof.squeeze()) ** 2
    return part1 - (part2 * part3) / part4


def mv_t_lr_partial_1_V_element(y, mat_mu, mat_d, mat_v, dof, i, j):
    # Would be nice to calculate only the necessary rows of Omega in the future maybe!
    # For part 2
    # zzT @ V
    # zzT[:, i, :] @ mat_v[:, :, j]
    # select the correct row of zzT before
    # sum(z * z[:, i], axis=-1)
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)

    # Deriviative for the log(det())
    part1 = np.sum(np.linalg.inv(precision)[:, i, :] * mat_v[:, :, j], axis=1)

    # Derivative for the second part
    deriv = -2 * np.sum(
        (y - mat_mu) * np.expand_dims((y[:, i] - mat_mu[:, i]), -1) * mat_v[:, :, j], -1
    )
    # Factor
    part2 = np.squeeze(k + dof)
    part3 = 2 * (Z + dof.squeeze())
    return part1 + (part2 / part3) * deriv


def mv_t_lr_partial_2_V_element(y, mat_mu, mat_d, mat_v, dof, i, j):
    d = y.shape[1]
    precision = mat_d + mat_v @ mat_v.swapaxes(-1, -2)
    cov = np.linalg.inv(precision)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)

    # second derivative for the log(det())
    sum1 = 0
    sum2 = 0
    for k, q in product(range(d), range(d)):
        sum1 += cov[:, i, i] * mat_v[:, q, j] * cov[:, q, k] * mat_v[:, k, j]
        sum2 += cov[:, i, q] * mat_v[:, q, j] * cov[:, i, k] * mat_v[:, k, j]
    part1 = cov[:, i, i] - sum1 - sum2

    # derivatives for the second part
    deriv1 = -2 * np.sum(
        (y - mat_mu) * np.expand_dims((y[:, i] - mat_mu[:, i]), -1) * mat_v[:, :, j], -1
    )
    deriv2 = 2 * (y - mat_mu)[:, i] ** 2

    part2 = d + dof.squeeze()
    part3 = (Z + dof.squeeze()) * deriv2 - deriv1**2
    part4 = 2 * (Z + dof.squeeze()) ** 2

    return part1 - (part2 * part3) / part4


def mv_t_lr_partial_1_dof(y, mat_mu, mat_d, mat_v, dof):
    k = y.shape[1]

    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)

    part1 = -(-dof * sp.digamma((k + dof) / 2) + k + dof * sp.digamma(dof / 2)) / (
        2 * dof
    )
    part2 = (Z * (k + dof)) / (2 * dof * (dof + Z)) - 1 / 2 * np.log((dof + Z) / dof)
    return part1 + part2


def mv_t_lr_partial_2_dof(y, mat_mu, mat_d, mat_v, dof):
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    Z = np.sum((y - mat_mu) * (precision @ (y - mat_mu)[..., None]).squeeze(), 1)
    part1 = (
        1
        / 4
        * (
            (2 * k) / (dof**2)
            + sp.polygamma(1, (dof + k) / 2)
            - sp.polygamma(1, dof / 2)
        )
    )
    part2 = (Z * (dof * Z - k * (2 * dof + Z))) / (2 * dof**2 * (dof + Z) ** 2)
    return part1 + part2



from typing import Dict

import numpy as np
import scipy.linalg as la
import scipy.stats as st

from ..base import Distribution, LinkFunction, MultivariateDistributionMixin
from ..links import Identity, Log, MatrixDiag, MatrixDiagTril
from ..types import ParameterShapes

# The MV gaussian distribution with modified Cholesky decomposition
# We use the notation T' D T = PRECISION
# Where T is a lower diagonal matrix and D is a diagonal matrix


class MultivariateNormalInverseModifiedCholesky(
    MultivariateDistributionMixin, Distribution
):
    """
    The multivariate normal (Gaussian) distribution parameterized by the modified Cholesky decomposition of the precision (inverse scale) matrix.

    In the modified Cholesky decomposition, the precision matrix \\( \\Omega \\) is written as:
    $$
    \\Omega = T^T D T
    $$
    where \\( T \\) is a lower triangular matrix with ones on the diagonal, and \\( D \\) is a diagonal matrix with positive entries.

    The PDF of the multivariate normal distribution is then:
    $$
    p(y \\mid \\mu, T, D) =
    |T| \\cdot |D|^{1/2} \\cdot (2\\pi)^{-k/2}
    \\exp\\left(-\\frac{1}{2} (y - \\mu)^T T^T D T (y - \\mu)\\right)
    $$
    where \\( k \\) is the dimensionality of the data, \\( \\mu \\) is the mean vector, \\( T \\) and \\( D \\) are the modified Cholesky factors of the precision matrix.
    """

    corresponding_gamlss: str = None
    parameter_names = {0: "mu", 1: "D", 2: "T"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (-np.inf, np.inf),
    }
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.multivariate_normal
    scipy_names = {"mu": "mean", "sigma": "cov"}
    parameter_shape = {
        0: ParameterShapes.VECTOR,
        1: ParameterShapes.DIAGONAL_MATRIX,
        2: ParameterShapes.LOWER_TRIANGULAR_MATRIX,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link_1: LinkFunction = MatrixDiag(Log()),
        scale_link_2: LinkFunction = MatrixDiagTril(Identity(), Identity()),
    ):
        super().__init__(
            links={
                0: loc_link,
                1: scale_link_1,
                2: scale_link_2,
            }
        )
        self._regularization_allowed = {0: False, 1: False, 2: True}

    def get_adr_regularization_distance(self, dim: int, param: int) -> float:
        if param in (0, 1):
            return None
        if param == 2:
            i, j = np.tril_indices(dim, k=-1)
            return np.abs(i - j)

    def get_regularization_size(self, dim: int) -> float:
        return dim - 1

    @staticmethod
    def fitted_elements(dim: int):
        return {0: dim, 1: dim, 2: int(dim * (dim - 1) // 2)}

    @staticmethod
    def index_flat_to_cube(k: int, d: int, param: int):
        if param == 0:
            return k
        if param == 1:
            return k, k
        if param == 2:
            # tril_indicies is row-wise
            # "inverted" triu_indicies is column-wise
            i, j = np.tril_indices(d, k=-1)
            return i[k], j[k]

    def set_theta_element(
        self, theta: Dict, value: np.ndarray, param: int, k: int
    ) -> Dict:
        """Sets an element of theta for parameter param and place k.

        !!! Note
            This will mutate `theta`!

        Args:
            theta (Dict): Current fitted $\theta$
            value (np.ndarray): Value to set
            param (int): Distribution parameter
            k (int): Flat element index $k$

        Returns:
            Dict: Theta where element (param, k) is set to value.
        """
        if param == 0:
            theta[param][:, k] = value
        if param == 1:
            theta[param][:, k, k] = value
        if param == 2:
            d = theta[0].shape[1]
            i, j = self.index_flat_to_cube(k, d, param)
            theta[param][:, i, j] = value
        return theta

    def theta_to_params(self, theta):
        mu = theta[0]
        d_mat = theta[1]
        t_mat = theta[2]
        return mu, d_mat, t_mat

    def set_initial_guess(
        self,
        y: np.ndarray,
        theta: Dict[int, np.ndarray],
        param: int,
    ) -> Dict[int, np.ndarray]:
        if param in (0, 2):
            return theta
        if param == 1:
            M = y.shape[0]
            residual = y - theta[0]
            var = np.var(residual, axis=0)
            shape = np.diag(var)
            _, d_inv, _ = la.ldl(shape, lower=True)
            d_mat = np.linalg.inv(d_inv)
            theta[1] = np.tile(d_mat, (M, 1, 1))
            return theta

    def dl1_dp1(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dp2(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dpp(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented.")

    def element_score(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl1_dp1(y=y, theta=theta, param=param, k=k)

    def element_hessian(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl2_dp2(y=y, theta=theta, param=param, k=k)

    def element_dl1_dp1(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip=False
    ):
        mu, d_mat, t_mat = self.theta_to_params(theta)
        if param == 0:
            deriv = _deriv1_mu(y, mu, d_mat, t_mat, i=k)
        if param == 1:
            deriv = _deriv1_dmat(y, mu, d_mat, t_mat, i=k)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            deriv = _deriv1_tmat(y, mu, d_mat, t_mat, i=i, j=j)
        return deriv

    def element_dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip=False
    ):
        mu, d_mat, t_mat = self.theta_to_params(theta)
        if param == 0:
            deriv = _deriv2_mu(y, mu, d_mat, t_mat, i=k)
        if param == 1:
            deriv = _deriv2_dmat(y, mu, d_mat, t_mat, i=k)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            deriv = _deriv2_tmat(y, mu, d_mat, t_mat, i=i, j=j)
        return deriv

    def element_link_function(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].link(y)
        if param >= 1:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_link(y, i=i, j=j)

    def element_link_function_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].link_derivative(y)
        if param >= 1:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_derivative(y, i=i, j=j)

    def element_link_function_second_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].link_second_derivative(y)
        if param >= 1:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_link_second_derivative(y, i=i, j=j)

    def element_link_inverse(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].inverse(y)
        if param >= 1:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_inverse(y, i=i, j=j)

    def element_link_inverse_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if param == 0:
            return self.links[param].inverse_derivative(y)
        if param >= 1:
            i, j = self.index_flat_to_cube(k=k, d=y.shape[1], param=param)
            return self.links[param].element_inverse_derivative(y, i=i, j=j)

    def initial_values(self, y, param=0):
        M = y.shape[0]
        mu = np.mean(y, axis=0)
        t_inv, d_inv, _ = la.ldl(np.cov(y, rowvar=False), lower=True)
        t_mat = np.linalg.inv(t_inv)
        d_mat = np.linalg.inv(d_inv)

        if param == 0:
            return np.tile(mu, (M, 1))
        if param == 1:
            return np.tile(d_mat, (M, 1, 1))
        if param == 2:
            return np.tile(t_mat, (M, 1, 1))

    def cube_to_flat(self, x: np.ndarray, param: int):
        if param == 0:
            return x
        if param == 1:
            k = x.shape[1]
            return x[:, range(k), range(k)]
        if param == 2:
            k = x.shape[1]
            # column wise indices
            i, j = np.tril_indices(k, k=-1)
            return x[:, i, j]

    def flat_to_cube(self, x: np.ndarray, param: int):
        if param == 0:
            return x
        if param == 1:
            k = x.shape[1]
            out = np.zeros((x.shape[0], k, k))
            out[:, range(k), range(k)] = x
            return out

        if param == 2:
            n, k = x.shape
            # The following conversion holds for upper diagonal matrices
            # We INCLUDE the diagonal!!
            # (D + 1) * D // 2 = k
            # (D + 1) * D = 2 * k
            # D^2 + D = 2 * k
            # ... Wolfram Alpha ...
            # D = 0.5 * (sqrt(8k + 1) - 1)
            d = int(1 / 2 * (np.sqrt(8 * k + 1) - 1))
            i, j = np.tril_indices(d + 1, k=-1)
            out = np.zeros((n, d + 1, d + 1))
            out[:, range(d + 1), range(d + 1)] = 1
            out[:, i, j] = x
            return out

    def param_conditional_likelihood(
        self, y: np.ndarray, theta: Dict, eta: np.ndarray, param: int
    ) -> np.ndarray:
        """Calulate the log-likelihood for (flat) eta for parameter (param)
        and theta for all other parameters.

        Args:
            y (np.ndarray): True values
            theta (Dict): Fitted theta.
            eta (np.ndarray): Fitted eta.
            param (int): Param for which we take eta.

        Returns:
            np.ndarray: Log-likelihood.
        """
        fitted = self.flat_to_cube(eta, param=param)
        fitted = self.link_inverse(fitted, param=param)
        # fitted_theta = {**theta, param: fitted_eta}
        return self.logpdf(y, theta={**theta, param: fitted})

    def theta_to_scipy(self, theta: Dict[int, np.ndarray]):
        mu, d_mat, t_mat = self.theta_to_params(theta)
        cov = np.linalg.inv(t_mat.swapaxes(-1, -2) @ d_mat @ t_mat)
        out = {
            "mean": mu,
            "cov": cov,
        }
        return out

    def cdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def ppf(self, q, theta):
        raise NotImplementedError("Not implemented")

    def rvs(self, size, theta):
        raise NotImplementedError("Not implemented")

    def logcdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def logpdf(self, y, theta):
        return _loglikelihood_gaussian_modchol(y, theta[0], theta[1], theta[2])

    def logpmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def calculate_conditional_initial_values(
        self, y: np.ndarray, theta: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        raise NotImplementedError("Not implemented")


def _loglikelihood_gaussian_modchol(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
) -> Dict[str, float]:
    k = y.shape[1]
    y_centered = y - mu
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    Z = np.sum(y_centered * (precision @ y_centered[..., None]).squeeze(), 1)
    part1 = -k / 2 * np.log(2 * np.pi)
    # part2 = 1 / 2 * np.sum(np.log(D_inv_mat[:, range(k), range(k)]), -1)
    part2 = 1 / 2 * np.log(np.linalg.det(D_inv))
    part3 = -1 / 2 * Z
    return part1 + part2 + part3


def _deriv1_mu(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    k = y.shape[1]
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    deriv_Z = -2 * np.sum(
        D_inv[:, range(i, k), range(i, k)]
        * Ty[:, range(i, k)]
        * T_mat[:, range(i, k), i],
        axis=1,
    )
    return -1 / 2 * deriv_Z


def _deriv2_mu(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    precision = np.swapaxes(T_mat, -2, -1) @ D_inv @ T_mat
    return -precision[:, i, i]


def _deriv1_dmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    deriv_Z = Ty[:, i] ** 2
    deriv_logdet = -1 / 2 * 1 / D_inv[:, i, i]
    return -deriv_logdet + (-1 / 2) * deriv_Z


def _deriv2_dmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    i: int = 0,
) -> np.ndarray:
    return -1 / 2 * 1 / D_inv[:, i, i] ** 2


def _deriv1_tmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    i: int = 0,
    j: int = 0,
) -> np.ndarray:
    y_centered = y - mu
    Ty = (T_mat @ y_centered[..., None]).squeeze(-1)
    deriv_Z = 2 * D_inv[:, i, i] * Ty[:, i] * y_centered[:, j]
    return (-1 / 2) * deriv_Z


def _deriv2_tmat(
    y: np.ndarray,
    mu: np.ndarray,
    D_inv: np.ndarray,
    T_mat: np.ndarray,
    i: int = 0,
    j: int = 0,
) -> np.ndarray:
    y_centered = y - mu
    deriv_Z = 2 * D_inv[:, i, i] * y_centered[:, j] ** 2
    return (-1 / 2) * deriv_Z
